# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import signal
from socket import gethostname
from typing import Any, Final
from collections.abc import Callable, Sequence

from xpra.os_util import gi_import
from xpra.util.env import envbool, envint, first_time
from xpra.util.io import get_proc_cmdline
from xpra.x11.common import Unmanageable
from xpra.gtk.gobject import one_arg_signal, n_arg_signal
from xpra.gtk.error import XError, xsync, xswallow, xlog
from xpra.codecs.image import ImageWrapper
from xpra.platform.posix.proc import get_parent_pid
from xpra.x11.bindings.window import X11WindowBindings, constants, SHAPE_KIND
from xpra.x11.bindings.res import ResBindings
from xpra.x11.bindings.send_wm import send_wm_delete_window
from xpra.x11.models.model_stub import WindowModelStub
from xpra.x11.gtk.composite import CompositeHelper
from xpra.x11.gtk.prop import prop_get, prop_set, prop_del, prop_type_get, PYTHON_TYPES
from xpra.x11.gtk.bindings import add_event_receiver, remove_event_receiver, get_pywindow
from xpra.log import Logger

log = Logger("x11", "window")
metalog = Logger("x11", "window", "metadata")
shapelog = Logger("x11", "window", "shape")
grablog = Logger("x11", "window", "grab")
framelog = Logger("x11", "window", "frame")
geomlog = Logger("x11", "window", "geometry")

GObject = gi_import("GObject")
GLib = gi_import("GLib")

X11Window = X11WindowBindings()

XRes = ResBindings()
if not XRes.check_xres():
    log.warn("Warning: X Resource Extension missing or too old")
    XRes = None

if get_parent_pid is None:
    log("proc.get_parent_pid is not available")

FORCE_QUIT = envbool("XPRA_FORCE_QUIT", True)
XSHAPE = envbool("XPRA_XSHAPE", True)
FRAME_EXTENTS = envbool("XPRA_FRAME_EXTENTS", True)
OPAQUE_REGION = envbool("XPRA_OPAQUE_REGION", True)
DELETE_DESTROY = envbool("XPRA_DELETE_DESTROY", False)
DELETE_KILL_PID = envbool("XPRA_DELETE_KILL_PID", True)
DELETE_XKILL = envbool("XPRA_DELETE_XKILL", True)

CurrentTime: Final[int] = constants["CurrentTime"]

# Re-stacking:
Above = 0
Below = 1
TopIf = 2
BottomIf = 3
Opposite = 4
RESTACKING_STR: dict[int, str] = {
    Above: "Above",
    Below: "Below",
    TopIf: "TopIf",
    BottomIf: "BottomIf",
    Opposite: "Opposite",
}

# grab stuff:
NotifyNormal: Final[int] = constants["NotifyNormal"]
NotifyGrab: Final[int] = constants["NotifyGrab"]
NotifyUngrab: Final[int] = constants["NotifyUngrab"]
NotifyWhileGrabbed: Final[int] = constants["NotifyWhileGrabbed"]
NotifyNonlinearVirtual: Final[int] = constants["NotifyNonlinearVirtual"]
GRAB_CONSTANTS: dict[int, str] = {
    NotifyNormal: "NotifyNormal",
    NotifyGrab: "NotifyGrab",
    NotifyUngrab: "NotifyUngrab",
    NotifyWhileGrabbed: "NotifyWhileGrabbed",
}
DETAIL_CONSTANTS: dict[int, str] = {}
for dconst in (
        "NotifyAncestor", "NotifyVirtual", "NotifyInferior",
        "NotifyNonlinear", "NotifyNonlinearVirtual", "NotifyPointer",
        "NotifyPointerRoot", "NotifyDetailNone",
):
    DETAIL_CONSTANTS[constants[dconst]] = dconst
grablog("pointer grab constants: %s", GRAB_CONSTANTS)
grablog("detail constants: %s", DETAIL_CONSTANTS)

# these properties are not handled, and we don't want to spam the log file
# whenever an app decides to change them:
PROPERTIES_IGNORED = [x for x in os.environ.get("XPRA_X11_PROPERTIES_IGNORED", "").split(",") if x]
# make it easier to debug property changes, just add them here:
# ie: {"WM_PROTOCOLS" : ["atom"]}
X11_PROPERTIES_DEBUG: dict[str, Any] = {}
PROPERTIES_DEBUG = [
    prop_debug.strip()
    for prop_debug in os.environ.get("XPRA_WINDOW_PROPERTIES_DEBUG", "").split(",")
]
X11PROPERTY_SYNC = envbool("XPRA_X11PROPERTY_SYNC", True)
X11PROPERTY_SYNC_BLOCKLIST = os.environ.get("XPRA_X11PROPERTY_SYNC_BLOCKLIST",
                                            "_GTK,WM_,_NET,Xdnd").split(",")
SHAPE_DELAY = envint("XPRA_SHAPE_DELAY", 100)


def sanestr(s: str) -> str:
    return s.strip("\0").replace("\0", " ")


class CoreX11WindowModel(WindowModelStub):
    """
        The utility superclass for all GTK / X11 window models,
        it wraps an X11 window (the "client-window").
        Defines the common properties and signals,
        sets up the composite helper, so we get the damage events.
        The x11_property_handlers sync X11 window properties into Python objects,
        the py_property_handlers do it in the other direction.
    """
    __common_properties__ = {
        # the X11 window id
        "xid": (
            GObject.TYPE_INT,
            "X11 window id", "",
            -1, 65535, -1,
            GObject.ParamFlags.READABLE,
        ),
        "parent": (
            GObject.TYPE_PYOBJECT,
            "parent window id", "",
            GObject.ParamFlags.READABLE,
        ),
        # FIXME: this is an ugly virtual property
        "geometry": (
            GObject.TYPE_PYOBJECT,
            "current coordinates (x, y, w, h, border) for the window", "",
            GObject.ParamFlags.READABLE,
        ),
        # bits per pixel
        "depth": (
            GObject.TYPE_INT,
            "window bit depth", "",
            -1, 64, -1,
            GObject.ParamFlags.READABLE,
        ),
        # if the window depth is 32 bit
        "has-alpha": (
            GObject.TYPE_BOOLEAN,
            "Does the window use transparency", "",
            False,
            GObject.ParamFlags.READABLE,
        ),
        # from WM_CLIENT_MACHINE
        "client-machine": (
            GObject.TYPE_PYOBJECT,
            "Host where client process is running", "",
            GObject.ParamFlags.READABLE,
        ),
        # from XResGetClientPid
        "pid": (
            GObject.TYPE_INT,
            "PID of owning process", "",
            -1, 65535, -1,
            GObject.ParamFlags.READABLE,
        ),
        "ppid": (
            GObject.TYPE_INT,
            "PID of parent process", "",
            -1, 65535, -1,
            GObject.ParamFlags.READABLE,
        ),
        # from _NET_WM_PID
        "wm-pid": (
            GObject.TYPE_INT,
            "PID of owning process", "",
            -1, 65535, -1,
            GObject.ParamFlags.READABLE,
        ),
        # from _NET_WM_NAME or WM_NAME
        "title": (
            GObject.TYPE_PYOBJECT,
            "Window title (unicode or None)", "",
            GObject.ParamFlags.READABLE,
        ),
        # from WM_WINDOW_ROLE
        "role": (
            GObject.TYPE_PYOBJECT,
            "The window's role (ICCCM session management)", "",
            GObject.ParamFlags.READABLE,
        ),
        # from WM_PROTOCOLS via XGetWMProtocols
        "protocols": (
            GObject.TYPE_PYOBJECT,
            "Supported WM protocols", "",
            GObject.ParamFlags.READABLE,
        ),
        # from WM_COMMAND
        "command": (
            GObject.TYPE_PYOBJECT,
            "Command used to start or restart the client", "",
            GObject.ParamFlags.READABLE,
        ),
        # from WM_CLASS via getClassHint
        "class-instance": (
            GObject.TYPE_PYOBJECT,
            "Classic X 'class' and 'instance'", "",
            GObject.ParamFlags.READABLE,
        ),
        # ShapeNotify events will populate this using XShapeQueryExtents
        "shape": (
            GObject.TYPE_PYOBJECT,
            "Window XShape data", "",
            GObject.ParamFlags.READABLE,
        ),
        # synced to "_NET_FRAME_EXTENTS"
        "frame": (
            GObject.TYPE_PYOBJECT,
            "Size of the window frame, as per _NET_FRAME_EXTENTS", "",
            GObject.ParamFlags.READWRITE,
        ),
        # synced to "_NET_WM_ALLOWED_ACTIONS"
        "allowed-actions": (
            GObject.TYPE_PYOBJECT,
            "Supported WM actions", "",
            GObject.ParamFlags.READWRITE,
        ),
        # synced to "_NET_WM_OPAQUE_REGION"
        "opaque-region": (
            GObject.TYPE_PYOBJECT,
            "Compositor can assume that there is no transparency for this region", "",
            GObject.ParamFlags.READWRITE,
        ),
    }

    __common_signals__ = {
        # signals we emit:
        "unmanaged": one_arg_signal,
        "restack": n_arg_signal(2),
        "initiate-moveresize": one_arg_signal,
        "grab": one_arg_signal,
        "ungrab": one_arg_signal,
        "bell": one_arg_signal,
        "client-contents-changed": one_arg_signal,
        "motion": one_arg_signal,
        # x11 events we catch (and often re-emit as something else):
        "x11-property-notify-event": one_arg_signal,
        "x11-xkb-event": one_arg_signal,
        "x11-shape-event": one_arg_signal,
        "x11-configure-event": one_arg_signal,
        "x11-unmap-event": one_arg_signal,
        "x11-client-message-event": one_arg_signal,
        "x11-focus-in-event": one_arg_signal,
        "x11-focus-out-event": one_arg_signal,
        "x11-motion-event": one_arg_signal,
        "x11-property-changed": one_arg_signal,
    }

    # things that we expose:
    _property_names = [
        "xid", "depth", "has-alpha",
        "parent",
        "client-machine", "pid", "ppid", "wm-pid",
        "title", "role",
        "command", "shape",
        "class-instance", "protocols",
        "opaque-region",
    ]
    # exposed and changing (should be watched for notify signals):
    _dynamic_property_names = [
        "title", "command", "shape", "class-instance", "protocols", "opaque-region",
    ]
    # should not be exported to the clients:
    _internal_property_names = [
        "frame", "allowed-actions",
    ]
    _initial_x11_properties = [
        "_NET_WM_PID", "WM_CLIENT_MACHINE",
        # _NET_WM_NAME is redundant, as it calls the same handler as "WM_NAME"
        "WM_NAME", "_NET_WM_NAME",
        "WM_PROTOCOLS", "WM_CLASS", "WM_WINDOW_ROLE",
        "_NET_WM_OPAQUE_REGION",
        "WM_COMMAND",
    ]
    _DEFAULT_NET_WM_ALLOWED_ACTIONS: list[str] = []
    _MODELTYPE = "Core"
    _scrub_x11_properties = [
        "WM_STATE",
        # "_NET_WM_STATE",    # "..it should leave the property in place when it is shutting down"
        "_NET_FRAME_EXTENTS", "_NET_WM_ALLOWED_ACTIONS",
    ]

    def __init__(self, xid: int):
        super().__init__()
        if not isinstance(xid, int):
            raise TypeError(f"xid must be an int, not a {type(xid)}")
        log("new window %#x", xid)
        self.xid: int = xid
        self._composite: CompositeHelper | None = None
        self._damage_forward_handle = None
        self._setup_done = False
        self._kill_count = 0
        self._shape_timer = 0
        self._shape_timer_serial = 0

    def __repr__(self) -> str:  # pylint: disable=arguments-differ
        try:
            classname = type(self).__name__
            return f"{classname}({self.xid:x})"
        except AttributeError:
            return repr(self)

    #########################################
    # Setup and teardown
    #########################################

    def call_setup(self) -> None:
        """
            Call this method to prepare the window:
            * makes sure it still exists
              (by querying its geometry which may raise an XError)
            * setup composite redirection
            * calls `setup`
            The difficulty comes from X11 errors and synchronization:
            we want to catch errors and undo what we've done.
            The mix of GTK and pure-X11 calls is not helping.
        """
        try:
            with xsync:
                geom = X11Window.geometry_with_border(self.xid)
                if geom is None:
                    raise Unmanageable(f"window {self.xid:x} disappeared already")
                self._internal_set_property("geometry", geom[:4])
                self._read_initial_X11_properties()
        except XError as e:
            log("failed to manage %#x", self.xid, exc_info=True)
            raise Unmanageable(e) from e
        add_event_receiver(self.xid, self)
        # Keith Packard says that composite state is undefined following a
        # reparent, so I'm not sure doing this here in the superclass,
        # before we reparent, actually works... let's wait and see.
        try:
            self._composite = CompositeHelper(self.xid)
            with xsync:
                self._composite.setup()
                if X11Window.displayHasXShape():
                    X11Window.XShapeSelectInput(self.xid)
        except Exception as e:
            remove_event_receiver(self.xid, self)
            log("%s %#x does not support compositing: %s", self._MODELTYPE, self.xid, e)
            with xswallow:
                self._composite.destroy()
            self._composite = None
            if isinstance(e, Unmanageable):
                raise
            raise Unmanageable(e) from e
        # compositing is now enabled,
        # from now on we must call setup_failed to clean things up
        self._managed = True
        try:
            with xsync:
                self.setup()
        except XError as e:
            log("failed to setup %#x", self.xid, exc_info=True)
            try:
                with xsync:
                    self.setup_failed(e)
            except Exception:
                log.error(f"Error in cleanup handler for window {self.xid:x}", exc_info=True)
            raise Unmanageable(e) from None
        self._setup_done = True

    def setup_failed(self, e) -> None:
        log("cannot manage %s %#x: %s", self._MODELTYPE, self.xid, e)
        self.do_unmanaged(False)

    def setup(self) -> None:
        # Start listening for important events.
        X11Window.addDefaultEvents(self.xid)
        self._damage_forward_handle = self._composite.connect("contents-changed", self._forward_contents_changed)
        self._setup_property_sync()

    def unmanage(self, exiting=False) -> None:
        if self._managed:
            self.emit("unmanaged", exiting)

    def do_unmanaged(self, wm_exiting) -> None:
        if not self._managed:
            return
        self._managed = False
        log("%s.do_unmanaged(%s) damage_forward_handle=%s, composite=%s",
            self._MODELTYPE, wm_exiting, self._damage_forward_handle, self._composite)
        remove_event_receiver(self.xid, self)
        self.managed_disconnect()
        if self._composite:
            if self._damage_forward_handle:
                self._composite.disconnect(self._damage_forward_handle)
                self._damage_forward_handle = None
            self._composite.destroy()
            self._composite = None
            self._scrub_x11()
            self.xid = 0

    #########################################
    # Damage / Composite
    #########################################

    def acknowledge_changes(self) -> None:
        if not self._managed:
            return
        c = self._composite
        if not c:
            raise RuntimeError("composite window destroyed outside the UI thread?")
        c.acknowledge_changes()

    def _forward_contents_changed(self, _obj, event) -> None:
        if self._managed:
            self.emit("client-contents-changed", event)

    def uses_xshm(self) -> bool:
        c = self._composite
        return bool(c) and c.has_xshm()

    def get_image(self, x: int, y: int, width: int, height: int) -> ImageWrapper:
        return self._composite.get_image(x, y, width, height)

    def _setup_property_sync(self) -> None:
        metalog("setup_property_sync()")
        # python properties which trigger an X11 property to be updated:
        for prop, cb in self._py_property_handlers.items():
            self.connect(f"notify::{prop}", cb)
        # initial sync:
        for cb in self._py_property_handlers.values():
            cb(self)
        # this one is special, and overridden in BaseWindow too:
        self.managed_connect("notify::protocols", self._update_can_focus)

    def _update_can_focus(self, *_args) -> None:
        can_focus = "WM_TAKE_FOCUS" in self.get_property("protocols")
        self._updateprop("can-focus", can_focus)

    def _read_initial_X11_properties(self) -> None:
        """ This is called within an XSync context,
            so that X11 calls can raise XErrors,
            pure GTK calls are not allowed. (they would trap the X11 error and crash!)
            Calling _updateprop is safe, because setup has not completed yet,
            so the property update will not fire notify()
        """
        metalog("read_initial_X11_properties() core")
        # immutable ones:
        depth = X11Window.get_depth(self.xid)
        pid = XRes.get_pid(self.xid) if XRes else -1
        ppid = get_parent_pid(pid) if pid and get_parent_pid else 0
        parent = X11Window.getParent(self.xid)
        if parent == X11Window.get_root_xid():
            parent = 0
        metalog("initial X11 properties: xid=%#x, parent=%#x, depth=%i, pid=%i, ppid=%i",
                self.xid, parent, depth, pid, ppid)
        self._updateprop("depth", depth)
        self._updateprop("xid", self.xid)
        self._updateprop("pid", pid)
        self._updateprop("ppid", ppid)
        self._updateprop("has-alpha", depth == 32)
        self._updateprop("allowed-actions", self._DEFAULT_NET_WM_ALLOWED_ACTIONS)
        self._updateprop("shape", self._read_xshape())
        self._updateprop("parent", get_pywindow(parent))
        # note: some of those are technically mutable,
        # but we don't export them as "dynamic" properties, so this won't be propagated
        # maybe we want to catch errors parsing _NET_WM_ICON ?
        metalog("initial X11_properties: querying %s", self._initial_x11_properties)
        # to make sure we don't call the same handler twice which is pointless
        # (the same handler may handle more than one X11 property)
        handlers = set()
        for mutable in self._initial_x11_properties:
            handler = self._x11_property_handlers.get(mutable)
            if not handler:
                log.error(f"Error: unknown initial X11 property: {mutable!r}")
            elif handler not in handlers:
                handlers.add(handler)
                try:
                    handler(self)
                except XError:
                    log("handler %s failed", handler, exc_info=True)
                    # these will be caught in call_setup()
                    raise
                except Exception:
                    # try to continue:
                    log.error("Error parsing initial property {mutable!r}", exc_info=True)

    def _scrub_x11(self) -> None:
        metalog("scrub_x11() x11 properties=%s", self._scrub_x11_properties)
        if not self._scrub_x11_properties:
            return
        with xswallow:
            for prop in self._scrub_x11_properties:
                X11Window.XDeleteProperty(self.xid, prop)

    #########################################
    # XShape
    #########################################

    def read_xshape(self):
        shapelog("read_xshape()")
        self._shape_timer = 0
        cur_shape = self.get_property("shape") or {}
        # remove serial before comparing dicts:
        cur_shape.pop("serial", None)
        # read new xshape:
        with xlog:
            # should we pass the x and y offsets here?
            # v = self._read_xshape(event.x, event.y)
            v = self._read_xshape()
            if cur_shape == v:
                shapelog("xshape unchanged")
                return
            v["serial"] = int(self._shape_timer_serial)
            shapelog("xshape updated with serial %#x", self._shape_timer_serial)
            self._internal_set_property("shape", v)

    def _read_xshape(self, x: int = 0, y: int = 0) -> dict[str, Any]:
        if not X11Window.displayHasXShape() or not XSHAPE:
            return {}
        extents = X11Window.XShapeQueryExtents(self.xid)
        if not extents:
            shapelog("read_shape for window %#x: no extents", self.xid)
            return {}
        # w,h = X11Window.getGeometry(xid)[2:4]
        shapelog("read_shape for window %#x: extents=%s", self.xid, extents)
        bextents = extents[0]
        cextents = extents[1]
        if bextents[0] == 0 and cextents[0] == 0:
            shapelog("read_shape for window %#x: none enabled", self.xid)
            return {}
        v = {
            "x": x,
            "y": y,
            "Bounding.extents": bextents,
            "Clip.extents": cextents,
        }
        for kind, kind_name in SHAPE_KIND.items():  # @UndefinedVariable
            rectangles = X11Window.XShapeGetRectangles(self.xid, kind)
            v[kind_name + ".rectangles"] = rectangles
        shapelog("_read_shape()=%s", v)
        return v

    ################################
    # Property reading
    ################################

    def get_dimensions(self):
        # just extracts the size from the geometry:
        return self.get_property("geometry")[2:4]

    def get_geometry(self):
        return self.get_property("geometry")[:4]

    #########################################
    # Python objects synced to X11 properties
    #########################################

    def _sync_allowed_actions(self, *_args) -> None:
        actions = self.get_property("allowed-actions") or []
        metalog("sync_allowed_actions: setting _NET_WM_ALLOWED_ACTIONS=%s on %#x", actions, self.xid)
        with xswallow:
            self.prop_set("_NET_WM_ALLOWED_ACTIONS", ["atom"], actions)

    def _handle_frame_changed(self, *_args) -> None:
        # legacy name for _sync_frame() called from Wm
        self._sync_frame()

    def _sync_frame(self, *_args) -> None:
        if not FRAME_EXTENTS:
            return
        v = self.get_property("frame")
        framelog("sync_frame: frame(%#x)=%s", self.xid, v)
        if not v and (not self.is_OR() and not self.is_tray()):
            v = self.root_prop_get("DEFAULT_NET_FRAME_EXTENTS", ["u32"])
        if not v:
            # default for OR, or if we don't have any other value:
            v = (0, 0, 0, 0)
        framelog("sync_frame: setting _NET_FRAME_EXTENTS=%s on %#x", v, self.xid)
        with xswallow:
            self.prop_set("_NET_FRAME_EXTENTS", ["u32"], v)

    _py_property_handlers: dict[str, Callable] = {
        "allowed-actions": _sync_allowed_actions,
        "frame": _sync_frame,
    }

    #########################################
    # X11 properties synced to Python objects
    #########################################

    def prop_get(self, key, ptype, ignore_errors: bool | None = None, raise_xerrors=False) -> object:
        """
            Get an X11 property from the client window,
            using the automatic type conversion code from prop.py
            Ignores property errors during setup_client.
        """
        if ignore_errors is None and (not self._setup_done or not self._managed):
            ignore_errors = True
        return prop_get(self.xid, key, ptype, ignore_errors=bool(ignore_errors), raise_xerrors=raise_xerrors)

    def prop_del(self, key: str) -> None:
        prop_del(self.xid, key)

    def prop_set(self, key: str, ptype, value) -> None:
        prop_set(self.xid, key, ptype, value)

    def root_prop_get(self, key: str, ptype, ignore_errors=True) -> object:
        return prop_get(X11Window.get_root_xid(), key, ptype, ignore_errors=ignore_errors)

    def root_prop_set(self, key: str, ptype, value) -> None:
        prop_set(X11Window.get_root_xid(), key, ptype, value)

    def do_x11_property_notify_event(self, event) -> None:
        # X11: PropertyNotify
        assert event.window == self.xid
        self._handle_property_change(str(event.atom))

    def _handle_property_change(self, name: str) -> None:
        # ie: _handle_property_change("_NET_WM_NAME")
        metalog("Property changed on %#x: %s", self.xid, name)
        x11proptype = X11_PROPERTIES_DEBUG.get(name)
        if x11proptype is not None:
            metalog.info("%s=%s", name, self.prop_get(name, x11proptype, True, False))
        if name in PROPERTIES_IGNORED:
            return
        if X11PROPERTY_SYNC and not any(name.startswith(x) for x in X11PROPERTY_SYNC_BLOCKLIST):
            try:
                with xsync:
                    prop_type = prop_type_get(self.xid, name)
                    metalog("_handle_property_change(%s) property type=%s", name, prop_type)
                    if prop_type:
                        dtype, dformat = prop_type
                        ptype = PYTHON_TYPES.get(dtype, "")
                        if ptype:
                            value = self.prop_get(name, ptype, ignore_errors=True)
                            if value is None:
                                # retry using scalar type:
                                scalar = (ptype,)
                                value = self.prop_get(name, scalar, ignore_errors=True)
                            metalog("_handle_property_change(%s) value=%s", name, value)
                            if value is not None:
                                self.emit("x11-property-changed", (name, ptype, dformat, value))
                                return
            except Exception:
                metalog("_handle_property_change(%s)", name, exc_info=True)
            self.emit("x11-property-changed", (name, "", 0, ""))
        handler = self._x11_property_handlers.get(name)
        if handler:
            try:
                with xsync:
                    handler(self)
            except XError as e:
                log("_handle_property_change", exc_info=True)
                log.error("Error processing property change for '%s'", name)
                log.error(" on window %#x", self.xid)
                log.estr(e)

    # specific properties:
    def _handle_pid_change(self) -> None:
        pid = self.prop_get("_NET_WM_PID", "u32") or -1
        metalog("_NET_WM_PID=%s", pid)
        self._updateprop("wm-pid", pid)

    def _handle_client_machine_change(self) -> None:
        client_machine = self.prop_get("WM_CLIENT_MACHINE", "latin1") or ""
        metalog("WM_CLIENT_MACHINE=%s", client_machine)
        self._updateprop("client-machine", client_machine)

    def _handle_wm_name_change(self) -> None:
        name = self.prop_get("_NET_WM_NAME", "utf8", True)
        metalog("_NET_WM_NAME=%s", name)
        if name is None:
            name = self.prop_get("WM_NAME", "latin1", True) or ""
            metalog("WM_NAME=%s", name)
        if self._updateprop("title", sanestr(str(name) or "")):
            metalog("wm_name changed")

    def _handle_role_change(self) -> None:
        role = self.prop_get("WM_WINDOW_ROLE", "latin1") or ""
        metalog("WM_WINDOW_ROLE=%s", role)
        self._updateprop("role", role)

    def _handle_protocols_change(self) -> None:
        with xsync:
            protocols = X11Window.XGetWMProtocols(self.xid)
        metalog("WM_PROTOCOLS=%s", protocols)
        self._updateprop("protocols", protocols)

    def _handle_command_change(self) -> None:
        command = self.prop_get("WM_COMMAND", "latin1")
        metalog("WM_COMMAND=%s", command)
        if command:
            command = command.strip("\0")
        else:
            pid = self.get_property("pid")
            command = " ".join(get_proc_cmdline(pid))
        self._updateprop("command", command)

    def _handle_class_change(self) -> None:
        class_instance = X11Window.getClassHint(self.xid)
        if class_instance:
            class_instance = tuple(v.decode("latin1") for v in class_instance)
        metalog("WM_CLASS=%s", class_instance)
        self._updateprop("class-instance", class_instance or ())

    def _handle_opaque_region_change(self) -> None:
        rectangles = []
        v: Sequence[int] = tuple(self.prop_get("_NET_WM_OPAQUE_REGION", ["u32"]) or [])
        if OPAQUE_REGION and len(v) % 4 == 0:
            while v:
                rvalues = tuple((coord if coord < 2 ** 32 else -1) for coord in v[:4])
                rectangles.append(rvalues)
                v = v[4:]
        metalog("_NET_WM_OPAQUE_REGION(%s)=%s (OPAQUE_REGION=%s)", v, rectangles, OPAQUE_REGION)
        self._updateprop("opaque-region", tuple(rectangles))

    # these handlers must not generate X11 errors (must use XSync)
    _x11_property_handlers: dict[str, Callable] = {
        "_NET_WM_PID": _handle_pid_change,
        "WM_CLIENT_MACHINE": _handle_client_machine_change,
        "WM_NAME": _handle_wm_name_change,
        "_NET_WM_NAME": _handle_wm_name_change,
        "WM_WINDOW_ROLE": _handle_role_change,
        "WM_PROTOCOLS": _handle_protocols_change,
        "WM_COMMAND": _handle_command_change,
        "WM_CLASS": _handle_class_change,
        "_NET_WM_OPAQUE_REGION": _handle_opaque_region_change,
    }

    #########################################
    # X11 Events
    #########################################

    def do_x11_unmap_event(self, event) -> None:
        log("do_x11_unmap_event(%s) xid=%s, ", event, self.xid)
        self.unmanage()

    def do_x11_destroy_event(self, event) -> None:
        log("do_x11_destroy_event(%s) xid=%s, ", event, self.xid)
        if event.delivered_to == self.xid:
            # This is somewhat redundant with the unmap signal, because if you
            # destroy a mapped window, then a UnmapNotify is always generated.
            # However, this allows us to catch the destruction of unmapped
            # ("iconified") windows, and also catch any mistakes we might have
            # made with unmap heuristics.  I love the smell of XDestroyWindow in
            # the morning.  It makes for simple code:
            self.unmanage()

    def process_client_message_event(self, event) -> bool:
        # FIXME
        # Need to listen for:
        #   _NET_CURRENT_DESKTOP
        #   _NET_WM_PING responses
        # and maybe:
        #   _NET_RESTACK_WINDOW
        #   _NET_WM_STATE (more fully)
        if event.message_type == "_NET_CLOSE_WINDOW":
            log.info("_NET_CLOSE_WINDOW received by %s", self)
            self.request_close()
            return True
        if event.message_type == "_NET_REQUEST_FRAME_EXTENTS":
            framelog("_NET_REQUEST_FRAME_EXTENTS")
            self._handle_frame_changed()
            return True
        if event.message_type == "_NET_MOVERESIZE_WINDOW":
            # this is overridden in WindowModel, skipped everywhere else:
            geomlog("_NET_MOVERESIZE_WINDOW skipped on %s (data=%s)", self, event.data)
            return True
        if event.message_type == "":
            log("empty message type: %s", event)
            exid = event.window
            if first_time(f"empty-x11-window-message-type-{exid:x}"):
                log.warn(f"Warning: empty message type received for window {exid:x}")
                log.warn(" %s", event)
                log.warn(" further messages will be silently ignored")
            return True
        # not handled:
        return False

    def do_x11_configure_event(self, event) -> None:
        if not self._managed:
            return
        # shouldn't the border width always be 0?
        geom = (event.x, event.y, event.width, event.height)
        geomlog("CoreX11WindowModel.do_x11_configure_event(%s) xid=%#x, new geometry=%s",
                event, self.xid, geom)
        self._updateprop("geometry", geom)

    def do_x11_shape_event(self, event) -> None:
        shapelog("shape event: %s, kind=%s, timer=%s",
                 event, SHAPE_KIND.get(event.kind, event.kind), self._shape_timer)
        cur_shape = self.get_property("shape")
        if not event.shaped and not cur_shape:
            shapelog("no shape")
            return
        serial = event.serial
        if cur_shape and cur_shape.get("serial", 0) >= serial:
            shapelog("same or older xshape serial no: %#x (current=%#x)", serial, cur_shape.get("serial", 0))
            return
        if not self._shape_timer:
            self._shape_timer = GLib.timeout_add(SHAPE_DELAY, self.read_xshape)
        self._shape_timer_serial = max(self._shape_timer_serial, serial)

    def do_x11_xkb_event(self, event) -> None:
        # X11: XKBNotify
        log("WindowModel.do_x11_xkb_event(%r)", event)
        if event.subtype != "bell":
            log("WindowModel.do_x11_xkb_event(%r)", event, exc_info=True)
            log.error("Error: unknown xkb event type: %s", event.type)
            return
        event.window_model = self
        self.emit("bell", event)

    def do_x11_client_message_event(self, event) -> None:
        # X11: ClientMessage
        log("do_x11_client_message_event(%s)", event)
        if not event.data or len(event.data) != 5:
            log.warn("invalid event data: %s", event.data)
            return
        if not self.process_client_message_event(event):
            log.warn("do_x11_client_message_event(%s) not handled", event)

    def do_x11_focus_in_event(self, event) -> None:
        # X11: FocusIn
        grablog("focus_in_event(%s) mode=%s, detail=%s",
                event, GRAB_CONSTANTS.get(event.mode), DETAIL_CONSTANTS.get(event.detail, event.detail))
        if event.mode == NotifyNormal and event.detail == NotifyNonlinearVirtual:
            self.emit("restack", Above, 0)
        else:
            self.may_emit_grab(event)

    def do_x11_focus_out_event(self, event) -> None:
        # X11: FocusOut
        grablog("focus_out_event(%s) mode=%s, detail=%s",
                event, GRAB_CONSTANTS.get(event.mode), DETAIL_CONSTANTS.get(event.detail, event.detail))
        self.may_emit_grab(event)

    def may_emit_grab(self, event) -> None:
        if event.mode == NotifyGrab:
            grablog("emitting grab on %s", self)
            self.emit("grab", event)
        if event.mode == NotifyUngrab:
            grablog("emitting ungrab on %s", self)
            self.emit("ungrab", event)

    def do_x11_motion_event(self, event) -> None:
        self.emit("motion", event)

    ################################
    # Actions
    ################################

    def raise_window(self) -> None:
        X11Window.XRaiseWindow(self.xid)

    def set_active(self) -> None:
        self.root_prop_set("_NET_ACTIVE_WINDOW", "u32", self.xid)

    ################################
    # Killing clients:
    ################################

    def request_close(self) -> bool:
        if "WM_DELETE_WINDOW" in self.get_property("protocols"):
            self.send_delete()
            return True
        title = self.get_property("title")
        xid = self.get_property("xid")
        if FORCE_QUIT:
            log.info("window %#x ('%s') does not support WM_DELETE_WINDOW", xid, title)
            log.info(" using force quit")
            # You don't want to play ball?  Then no more Mr. Nice Guy!
            self.force_quit()
            return True
        log.warn("window %#x ('%s') cannot be closed,", xid, title)
        log.warn(" it does not support WM_DELETE_WINDOW")
        log.warn(" and FORCE_QUIT is disabled")
        return False

    def send_delete(self) -> None:
        with xswallow:
            now = X11Window.get_server_time(self.xid)
            send_wm_delete_window(self.xid, timestamp=now)

    def XKill(self) -> None:
        with xswallow:
            X11Window.XKillClient(self.xid)

    def force_quit(self) -> None:
        if DELETE_DESTROY:
            with xlog:
                X11Window.DestroyWindow(self.xid)
                return
        machine = self.get_property("client-machine")
        pid = self.get_property("pid")
        if pid <= 0:
            # we could fall back to _NET_WM_PID
            # but that would be unsafe
            log.warn("Warning: cannot terminate window %#x, no pid found", self.xid)
            if machine:
                log.warn(" WM_CLIENT_MACHINE=%s", machine)
            pid = self.get_property("wm-pid")
            if pid > 0:
                log.warn(" _NET_WM_PID=%s", pid)
        if pid == os.getpid():
            log.warn("Warning: force_quit is refusing to kill ourselves!")
            return
        localhost = gethostname()
        log("force_quit() pid=%s, machine=%s, localhost=%s", pid, machine, localhost)
        if DELETE_KILL_PID and pid and machine is not None and machine == localhost:
            if self._kill_count == 0:
                # first time around: just send a SIGINT and hope for the best
                try:
                    os.kill(pid, signal.SIGINT)
                except OSError as e:
                    log.warn("Warning: failed to kill(SIGINT) client with pid %s", pid)
                    log.warn(" %s", e)
            else:
                # the more brutal way: SIGKILL + XKill
                try:
                    os.kill(pid, signal.SIGKILL)
                except OSError as e:
                    log.warn("Warning: failed to kill(SIGKILL) client with pid %s", pid)
                    log.warn(" %s", e)
                self.XKill()
            self._kill_count += 1
            return
        if DELETE_XKILL:
            self.XKill()



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import os
import shlex
from shutil import which
from subprocess import PIPE, Popen
from collections.abc import Callable

from xpra.scripts.main import InitException, InitExit, shellquote, host_target_string
from xpra.net.bytestreams import ConnectionClosedException
from xpra.common import noop
from xpra.util.thread import start_thread
from xpra.exit_codes import ExitCode
from xpra.os_util import gi_import, WIN32, OSX, POSIX
from xpra.util.env import envbool, restore_script_env, get_saved_env
from xpra.log import Logger, is_debug_enabled

# pylint: disable=import-outside-toplevel

log = Logger("network", "ssh")
if log.is_debug_enabled():
    import logging
    logging.getLogger("paramiko").setLevel(logging.DEBUG)

MAGIC_QUOTES = envbool("XPRA_SSH_MAGIC_QUOTES", True)


def connect_failed(_message) -> None:
    # by the time ssh fails, we may have entered the gtk main loop
    # (and more than once thanks to the clipboard code..)
    if "gi.repository.Gtk" in sys.modules:
        gtk = gi_import("Gtk")
        gtk.main_quit()


def connect_to(display_desc, opts=None, debug_cb: Callable = noop, ssh_fail_cb=connect_failed):
    sshpass_command = None
    cmd = list(display_desc["full_ssh"])
    kwargs = {}
    env = display_desc.get("env")
    if env is None:
        env = get_saved_env()
    if display_desc.get("is_putty"):
        # special env used by plink:
        env = os.environ.copy()
        env["PLINK_PROTOCOL"] = "ssh"
    kwargs["stderr"] = sys.stderr
    try:
        if WIN32:
            from subprocess import (
                CREATE_NEW_PROCESS_GROUP, CREATE_NEW_CONSOLE, STARTUPINFO, STARTF_USESHOWWINDOW,  # @UnresolvedImport
            )
            startupinfo = STARTUPINFO()
            startupinfo.dwFlags |= STARTF_USESHOWWINDOW
            startupinfo.wShowWindow = 0     # aka win32.con.SW_HIDE
            flags = CREATE_NEW_PROCESS_GROUP | CREATE_NEW_CONSOLE
            kwargs |= {
                "startupinfo": startupinfo,
                "creationflags": flags,
                "stderr": PIPE,
            }
        elif not display_desc.get("exit_ssh", False) and not OSX:
            kwargs["start_new_session"] = True
        remote_xpra: list[str] = display_desc["remote_xpra"]
        assert remote_xpra
        socket_dir = display_desc.get("socket_dir")
        proxy_command: list[str] = display_desc["proxy_command"]      # ie: ["_proxy_start"]
        # display_as_args ie: ["--start=xterm", "--env=SSH_AGENT_UUID={uuid}", ":10"]
        display_as_args: list[str] = display_desc["display_as_args"]
        remote_cmd = ""
        for x in remote_xpra:
            check = "if" if not remote_cmd else "elif"
            if x == "xpra":
                # no absolute path, so use "command -v" to check that the command exists:
                pc = [f'{check} command -v "{x}" > /dev/null 2>&1; then']
            else:
                pc = [f'{check} [ -x {x} ]; then']
            pc += [x] + proxy_command + [shellquote(x) for x in display_as_args]
            if socket_dir:
                pc.append(f"--socket-dir={socket_dir}")
            remote_cmd += " ".join(pc)+";"
        remote_cmd += "else echo \"no run-xpra command found\"; exit 1; fi"
        # how many times we need to escape the remote command string
        # depends on how many times the ssh command is parsed
        nssh = sum(int(x == "ssh") for x in cmd)
        if nssh >= 2 and MAGIC_QUOTES:
            for _ in range(nssh):
                remote_cmd = shlex.quote(remote_cmd)
        else:
            remote_cmd = f"'{remote_cmd}'"
        cmd.append(f"sh -c {remote_cmd}")
        debug_cb(f"starting {cmd[0]} tunnel")
        # non-string arguments can make Popen choke,
        # instead of lazily converting everything to a string, we validate the command:
        for x in cmd:
            if not isinstance(x, str):
                raise InitException(f"argument is not a string: {x} ({type(x)}), found in command: {cmd}")
        password = display_desc.get("password")
        if password and not display_desc.get("is_putty", False):
            from xpra.platform.paths import get_sshpass_command
            sshpass_command = get_sshpass_command()
            if sshpass_command:
                # sshpass -e ssh ...
                cmd.insert(0, sshpass_command)
                cmd.insert(1, "-e")
                env["SSHPASS"] = password
                # the password will be used by ssh via sshpass,
                # don't try to authenticate again over the ssh-proxy connection,
                # which would trigger warnings if the server does not require
                # authentication over unix-domain-sockets:
                opts.password = None
                del display_desc["password"]

        kwargs["env"] = restore_script_env(env)

        log_fn = log.info if is_debug_enabled("ssh") else log.debug
        log_fn("executing ssh command: " + shlex.join(cmd))
        log_fn(f"{kwargs=}")
        executable = which(cmd[0])
        if not executable or not os.path.exists(executable):
            log.warn(f"Warning: ssh command {cmd[0]!r} not found!")
            log.warn(" trying to continue anyway..")
        child = Popen(cmd, stdin=PIPE, stdout=PIPE, **kwargs)
    except OSError as e:
        cmd_info = shlex.join(cmd)
        raise InitExit(ExitCode.SSH_FAILURE,
                       f"Error running ssh command {cmd_info!r}: {e}") from None

    def abort_test(action) -> None:
        """ if ssh dies, we don't need to try to read/write from its sockets """
        exitcode = child.poll()
        if exitcode is not None:
            had_connected = conn.input_bytecount > 0 or conn.output_bytecount > 0
            if had_connected:
                error_message = f"cannot {action} using SSH"
            else:
                error_message = "SSH connection failure"
            sshpass_error = None
            if sshpass_command:
                sshpass_error = {
                    1: "Invalid command line argument",
                    2: "Conflicting arguments given",
                    3: "General runtime error",
                    4: "Unrecognized response from ssh (parse error)",
                    5: "Invalid/incorrect password",
                    6: "Host public key is unknown. sshpass exits without confirming the new key.",
                }.get(exitcode)
                if sshpass_error:
                    error_message += f": {sshpass_error}"
            debug_cb(error_message)
            ssh_fail_cb(error_message)
            if "ssh_abort" not in display_desc:
                display_desc["ssh_abort"] = True
                if not had_connected:
                    log.error("Error: SSH connection to the xpra server failed")
                    if sshpass_error:
                        log.error(f" {sshpass_error}")
                    else:
                        log.error(" check your username, hostname, display number, firewall, etc")
                    display_name = display_desc["display_name"]
                    log.error(f" for server: {display_name}")
                else:
                    log.error(f"The SSH process has terminated with exit code {exitcode}")
                ssh_cmd_info = " ".join(display_desc["full_ssh"])
                log.error(" the command line used was:")
                log.error(f" {ssh_cmd_info}")
            raise ConnectionClosedException(error_message) from None

    def stop_tunnel() -> None:
        close_tunnel_pipes(child)
        if not display_desc.get("exit_ssh", False):
            # leave ssh running
            return
        try:
            if child.poll() is None:
                child.terminate()
        except OSError as term_err:
            log.error(f"Error trying to stop ssh tunnel process: {term_err}", exc_info=True)

    host: str = display_desc["host"]
    port: int = display_desc.get("port", 22)
    username: str = display_desc.get("username", "")
    display: str = display_desc.get("display", "")
    info = {
        "host": host,
        "port": port,
    }
    from xpra.net.bytestreams import TwoFileConnection
    conn = TwoFileConnection(child.stdin, child.stdout,
                             abort_test, target=(host, port),
                             socktype="ssh", close_cb=stop_tunnel, info=info)
    conn.endpoint = host_target_string("ssh", username, host, port, display)
    conn.timeout = 0            # taken care of by abort_test
    conn.process = (child, "ssh", cmd)
    if kwargs.get("stderr") == PIPE:

        start_thread(stderr_reader, "ssh-stderr-reader", args=(child,), daemon=True)
    return conn


def close_tunnel_pipes(child: Popen) -> None:
    if POSIX:
        # on posix, the tunnel may be shared with other processes
        # so don't kill it... which may leave it behind after use.
        # but at least make sure we close all the pipes:
        for name, fd in {
            "stdin": child.stdin,
            "stdout": child.stdout,
            "stderr": child.stderr,
        }.items():
            try:
                if fd:
                    fd.close()
            except Exception as fd_err:
                log.error(f"Error closing ssh tunnel {name}: {fd_err}")


def stderr_reader(child: Popen) -> None:
    errs: list[str] = []
    while child.poll() is None:
        try:
            # noinspection PyTypeChecker
            v: bytes = child.stderr.readline()
        except OSError:
            log("stderr_reader()", exc_info=True)
            break
        if not v:
            log("SSH EOF on stderr")
            break
        try:
            s = v.rstrip(b"\n\r").decode("utf8")
        except UnicodeDecodeError:
            s = str(v.rstrip(b"\n\r"))
        if s:
            errs.append(s)
    if errs:
        log.warn("SSH stderr:")
        for err in errs:
            log.warn(f" {err}")



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import monotonic
from libc.string cimport memset
from libc.stdint cimport uintptr_t
from libc.stdio cimport printf
from xpra.buffers.membuf cimport getbuf, buffer_context, MemBuf  # pylint: disable=syntax-error

from weakref import WeakValueDictionary
from typing import Any, Dict, Tuple
from collections.abc import Sequence
from threading import Event

from xpra.util.str_fn import csv
from xpra.util.objects import AtomicInteger, typedict
from xpra.codecs.image import ImageWrapper
from xpra.codecs.constants import VideoSpec
from xpra.codecs.nvidia.cuda.errors import cudacheck, get_error_name
from xpra.codecs.nvidia.cuda.context import get_default_device_context
from xpra.log import Logger
log = Logger("encoder", "nvdec")

#we can import pycuda safely here,
#because importing cuda/context will have imported it with the lock
from pycuda.driver import Memcpy2D, mem_alloc_pitch, memcpy_dtoh, Stream

cdef inline int roundup(int n, int m):
    return (n + m - 1) & ~(m - 1)

ctypedef int CUresult
ctypedef void* CUstream
ctypedef void* CUcontext
ctypedef void* CUvideodecoder
ctypedef void* CUvideoctxlock

ctypedef void *CUvideoparser
ctypedef long long CUvideotimestamp


DEF CUDA_ERROR_INVALID_IMAGE = 200

ctypedef struct rect:
    short left
    short top
    short right
    short bottom

cdef extern from "nvcuvid.h":
    ctypedef enum CUvideopacketflags:
        CUVID_PKT_ENDOFSTREAM   #Set when this is the last packet for this stream
        CUVID_PKT_TIMESTAMP     #Timestamp is valid
        CUVID_PKT_DISCONTINUITY #Set when a discontinuity has to be signalled
        CUVID_PKT_ENDOFPICTURE  #Set when the packet contains exactly one frame or one field
        CUVID_PKT_NOTIFY_EOS    #If this flag is set along with CUVID_PKT_ENDOFSTREAM, an additional (dummy)
                                #display callback will be invoked with null value of CUVIDPARSERDISPINFO which

    ctypedef struct CUSEIMESSAGE:
        unsigned char sei_message_type  #OUT: SEI Message Type
        unsigned char reserved[3]
        unsigned int sei_message_size   #OUT: SEI Message Size
    ctypedef struct CUVIDSEIMESSAGEINFO:
        void *pSEIData                  #OUT: SEI Message Data
        CUSEIMESSAGE *pSEIMessage       #OUT: SEI Message Info
        unsigned int sei_message_count  #OUT: SEI Message Count
        unsigned int picIdx             #OUT: SEI Message Pic Index

    ctypedef struct av1:
        unsigned char  operating_points_cnt
        unsigned char  reserved24_bits[3]
        unsigned short operating_points_idc[32]

    ctypedef struct CUVIDOPERATINGPOINTINFO:
        cudaVideoCodec codec
        av1 av1
        unsigned char CodecReserved[1024]

    ctypedef struct frame_rate:
        #frame_rate
        unsigned int numerator
        unsigned int denominator

    ctypedef struct CUVIDEOFORMAT:
        cudaVideoCodec codec                    #OUT: Compression format
        frame_rate frame_rate
        unsigned char progressive_sequence      #OUT: 0=interlaced, 1=progressive
        unsigned char bit_depth_luma_minus8     #OUT: high bit depth luma
        unsigned char bit_depth_chroma_minus8   #OUT: high bit depth chroma
        unsigned char min_num_decode_surfaces   #OUT: Minimum number of decode surfaces to be allocated for correct
        unsigned int coded_width                #OUT: coded frame width in pixels
        unsigned int coded_height               #OUT: coded frame height in pixels
        rect display_area
        cudaVideoChromaFormat chroma_format     #OUT:  Chroma format
        unsigned int bitrate                    #OUT: video bitrate (bps, 0=unknown)
        #display_aspect_ratio: int x, int y

    ctypedef struct CUVIDPARSERDISPINFO:
        int picture_index           #OUT: Index of the current picture                                                         */
        int progressive_frame       #OUT: 1 if progressive frame; 0 otherwise                                                  */
        int top_field_first         #OUT: 1 if top field is displayed first; 0 otherwise                                       */
        int repeat_first_field      #OUT: Number of additional fields (1=ivtc, 2=frame doubling, 4=frame tripling,
                                    #-1=unpaired field)                                                                        */
        CUvideotimestamp timestamp  #OUT: Presentation time stamp

    ctypedef int (*PFNVIDSEQUENCECALLBACK)(void *, CUVIDEOFORMAT *) except 0
    ctypedef int (*PFNVIDDECODECALLBACK)(void *, CUVIDPICPARAMS *) except 0
    ctypedef int (*PFNVIDDISPLAYCALLBACK)(void *, CUVIDPARSERDISPINFO *) except 0
    ctypedef int (*PFNVIDOPPOINTCALLBACK)(void *, CUVIDOPERATINGPOINTINFO*) except 0
    ctypedef int (*PFNVIDSEIMSGCALLBACK)(void *, CUVIDSEIMESSAGEINFO *) except 0

    ctypedef struct CUVIDEOFORMATEX:
        pass
    ctypedef struct CUVIDPARSERPARAMS:
        cudaVideoCodec CodecType                #IN: cudaVideoCodec_XXX
        unsigned int ulMaxNumDecodeSurfaces     #IN: Max # of decode surfaces (parser will cycle through these)
        unsigned int ulClockRate                #IN: Timestamp units in Hz (0=default=10000000Hz)
        unsigned int ulErrorThreshold           #IN: % Error threshold (0-100) for calling pfnDecodePicture (100=always
                                                #IN: call pfnDecodePicture even if picture bitstream is fully corrupted)
        unsigned int ulMaxDisplayDelay          #IN: Max display queue delay (improves pipelining of decode with display)
                                                #0=no delay (recommended values: 2..4)
        unsigned int bAnnexb                    #IN: AV1 annexB stream
        unsigned int uReserved                  #Reserved for future use - set to zero
        unsigned int uReserved1[4]              #IN: Reserved for future use - set to 0
        void *pUserData                         #IN: User data for callbacks
        PFNVIDSEQUENCECALLBACK pfnSequenceCallback  #IN: Called before decoding frames and/or whenever there is a fmt change
        PFNVIDDECODECALLBACK pfnDecodePicture       #IN: Called when a picture is ready to be decoded (decode order)
        PFNVIDDISPLAYCALLBACK pfnDisplayPicture     #IN: Called whenever a picture is ready to be displayed (display order)
        PFNVIDOPPOINTCALLBACK pfnGetOperatingPoint  #IN: Called from AV1 sequence header to get operating point of a AV1
                                                    #scalable bitstream
        PFNVIDSEIMSGCALLBACK pfnGetSEIMsg       #IN: Called when all SEI messages are parsed for particular frame
        void *pvReserved2[5]                    #Reserved for future use - set to NULL
        CUVIDEOFORMATEX *pExtVideoInfo          #IN: [Optional] sequence header data from system layer
    CUresult cuvidCreateVideoParser(CUvideoparser *pObj, CUVIDPARSERPARAMS *pParams)
    #CUresult cuvidDestroyVideoSource(CUvideosource obj)
    CUresult cuvidDestroyVideoParser(CUvideoparser obj)

    ctypedef struct CUVIDSOURCEDATAPACKET:
        unsigned long flags             #IN: Combination of CUVID_PKT_XXX flags
        unsigned long payload_size      #IN: number of bytes in the payload (may be zero if EOS flag is set)
        const unsigned char *payload    #IN: Pointer to packet payload data (may be NULL if EOS flag is set)
        CUvideotimestamp timestamp      #IN: Presentation time stamp (10MHz clock), only valid if
                                        #CUVID_PKT_TIMESTAMP flag is set
    CUresult cuvidParseVideoData(CUvideoparser obj, CUVIDSOURCEDATAPACKET *pPacket) nogil


cdef extern from "cuviddec.h":
    ctypedef enum cudaVideoCodec:
        cudaVideoCodec_MPEG1
        cudaVideoCodec_MPEG2
        cudaVideoCodec_MPEG4
        cudaVideoCodec_VC1
        cudaVideoCodec_H264
        cudaVideoCodec_JPEG
        cudaVideoCodec_H264_SVC
        cudaVideoCodec_H264_MVC
        cudaVideoCodec_HEVC
        cudaVideoCodec_VP8
        cudaVideoCodec_VP9
        cudaVideoCodec_AV1
        cudaVideoCodec_NumCodecs
        #Uncompressed YUV
        cudaVideoCodec_YUV420       #Y,U,V (4:2:0)
        cudaVideoCodec_YV12         #Y,V,U (4:2:0)
        cudaVideoCodec_NV12         #Y,UV  (4:2:0)
        cudaVideoCodec_YUYV         #YUYV/YUY2 (4:2:2)
        cudaVideoCodec_UYVY         #UYVY (4:2:2)
    ctypedef enum cudaVideoChromaFormat:
        cudaVideoChromaFormat_Monochrome
        cudaVideoChromaFormat_420
        cudaVideoChromaFormat_422
        cudaVideoChromaFormat_444
    ctypedef enum cudaVideoCreateFlags:
        cudaVideoCreate_Default         #Default operation mode: use dedicated video engines
        cudaVideoCreate_PreferCUDA      #Use CUDA-based decoder (requires valid vidLock object for multi-threading)
        cudaVideoCreate_PreferDXVA      #Go through DXVA internally if possible (requires D3D9 interop)
        cudaVideoCreate_PreferCUVID     #Use dedicated video engines directly

    ctypedef enum cuvidDecodeStatus:
        cuvidDecodeStatus_Invalid           #Decode status is not valid
        cuvidDecodeStatus_InProgress        #Decode is in progress
        cuvidDecodeStatus_Success           #Decode is completed without any errors
        #// 3 to 7 enums are reserved for future use
        cuvidDecodeStatus_Error             #Decode is completed with an error (error is not concealed)
        cuvidDecodeStatus_Error_Concealed   #Decode is completed with an error and error is concealed

    ctypedef struct CUVIDDECODECAPS:
        cudaVideoCodec          eCodecType              #IN: cudaVideoCodec_XXX
        cudaVideoChromaFormat   eChromaFormat           #IN: cudaVideoChromaFormat_XXX
        unsigned int            nBitDepthMinus8         #IN: The Value "BitDepth minus 8"
        unsigned int            reserved1[3]            #Reserved for future use - set to zero

        unsigned char           bIsSupported            #OUT: 1 if codec supported, 0 if not supported
        unsigned char           nNumNVDECs              #OUT: Number of NVDECs that can support IN params
        unsigned short          nOutputFormatMask       #OUT: each bit represents corresponding cudaVideoSurfaceFormat enum
        unsigned int            nMaxWidth               #OUT: Max supported coded width in pixels
        unsigned int            nMaxHeight              #OUT: Max supported coded height in pixels
        unsigned int            nMaxMBCount             #OUT: Max supported macroblock count
                                                        #CodedWidth*CodedHeight/256 must be <= nMaxMBCount
        unsigned short          nMinWidth               #OUT: Min supported coded width in pixels
        unsigned short          nMinHeight              #OUT: Min supported coded height in pixels
        unsigned char           bIsHistogramSupported   #OUT: 1 if Y component histogram output is supported, 0 if not
                                                        #Note: histogram is computed on original picture data before
                                                        #any post-processing like scaling, cropping, etc. is applied
        unsigned char           nCounterBitDepth        #OUT: histogram counter bit depth
        unsigned short          nMaxHistogramBins       #OUT: Max number of histogram bins
        unsigned int            reserved3[10]           #Reserved for future use - set to zero

    ctypedef enum cudaVideoDeinterlaceMode:
        cudaVideoDeinterlaceMode_Weave      #Weave both fields (no deinterlacing)
        cudaVideoDeinterlaceMode_Bob        #Drop one field
        cudaVideoDeinterlaceMode_Adaptive   #Adaptive deinterlacing

    ctypedef enum cudaVideoSurfaceFormat:
        cudaVideoSurfaceFormat_NV12                     #Semi-Planar YUV [Y plane followed by interleaved UV plane]
        cudaVideoSurfaceFormat_P016                     #16 bit Semi-Planar YUV [Y plane followed by interleaved UV plane].
                                                        #Can be used for 10 bit(6LSB bits 0), 12 bit (4LSB bits 0)
        cudaVideoSurfaceFormat_YUV444                   #Planar YUV [Y plane followed by U and V planes]
        cudaVideoSurfaceFormat_YUV444_16Bit             #16 bit Planar YUV [Y plane followed by U and V planes].
                                                        #Can be used for 10 bit(6LSB bits 0), 12 bit (4LSB bits 0)
    ctypedef struct CUVIDDECODECREATEINFO:
        unsigned long ulWidth               #IN: Coded sequence width in pixels
        unsigned long ulHeight              #IN: Coded sequence height in pixels
        unsigned long ulNumDecodeSurfaces   #IN: Maximum number of internal decode surfaces
        cudaVideoCodec CodecType            #IN: cudaVideoCodec_XXX
        cudaVideoChromaFormat ChromaFormat  #IN: cudaVideoChromaFormat_XXX
        unsigned long ulCreationFlags       #IN: Decoder creation flags (cudaVideoCreateFlags_XXX)
        unsigned long bitDepthMinus8        #IN: The value "BitDepth minus 8"
        unsigned long ulIntraDecodeOnly     #IN: Set 1 only if video has all intra frames (default value is 0). This will
                                            #optimize video memory for Intra frames only decoding. The support is limited
                                            #to specific codecs - H264, HEVC, VP9, the flag will be ignored for codecs which
                                            #are not supported. However decoding might fail if the flag is enabled in case
                                            #of supported codecs for regular bit streams having P and/or B frames.
        unsigned long ulMaxWidth            #Coded sequence max width in pixels used with reconfigure Decoder
        unsigned long ulMaxHeight           #IN: Coded sequence max height in pixels used with reconfigure Decoder
        unsigned long Reserved1             #Reserved for future use - set to zero
        rect display_area
        cudaVideoSurfaceFormat OutputFormat         #IN: cudaVideoSurfaceFormat_XXX
        cudaVideoDeinterlaceMode DeinterlaceMode    #IN: cudaVideoDeinterlaceMode_XXX
        unsigned long ulTargetWidth                 #IN: Post-processed output width (Should be aligned to 2)
        unsigned long ulTargetHeight                #IN: Post-processed output height (Should be aligned to 2)
        unsigned long ulNumOutputSurfaces           #IN: Maximum number of output surfaces simultaneously mapped
        CUvideoctxlock vidLock                      #IN: If non-NULL, context lock used for synchronizing ownership of
                                                    #the cuda context. Needed for cudaVideoCreate_PreferCUDA decode
        rect target_rect
        unsigned long enableHistogram               #IN: enable histogram output, if supported
        unsigned long Reserved2[4]                  #Reserved for future use - set to zero

    ctypedef struct CUVIDH264PICPARAMS:
        int log2_max_frame_num_minus4
        int pic_order_cnt_type
        int log2_max_pic_order_cnt_lsb_minus4
        int delta_pic_order_always_zero_flag
        int frame_mbs_only_flag
        int direct_8x8_inference_flag
        int num_ref_frames                          #NOTE: shall meet level 4.1 restrictions
        unsigned char residual_colour_transform_flag;
        unsigned char bit_depth_luma_minus8         #Must be 0 (only 8-bit supported)
        unsigned char bit_depth_chroma_minus8       #Must be 0 (only 8-bit supported)
        unsigned char qpprime_y_zero_transform_bypass_flag
        # PPS
        int entropy_coding_mode_flag
        int pic_order_present_flag
        int num_ref_idx_l0_active_minus1
        int num_ref_idx_l1_active_minus1
        int weighted_pred_flag
        int weighted_bipred_idc
        int pic_init_qp_minus26
        int deblocking_filter_control_present_flag
        int redundant_pic_cnt_present_flag
        int transform_8x8_mode_flag
        int MbaffFrameFlag
        int constrained_intra_pred_flag
        int chroma_qp_index_offset
        int second_chroma_qp_index_offset
        int ref_pic_flag
        int frame_num
        int CurrFieldOrderCnt[2]
        # DPB
        #CUVIDH264DPBENTRY dpb[16];          // List of reference frames within the DPB
        # Quantization Matrices (raster-order)
        unsigned char WeightScale4x4[6][16]
        unsigned char WeightScale8x8[2][64]
        # FMO/ASO
        unsigned char fmo_aso_enable
        unsigned char num_slice_groups_minus1
        unsigned char slice_group_map_type
        signed char pic_init_qs_minus26
        unsigned int slice_group_change_rate_minus1
        #union
        #{
        #unsigned long long slice_group_map_addr;
        #const unsigned char *pMb2SliceGroupMap;
        #} fmo;
        unsigned int  Reserved[12]

    ctypedef union CUVIDCodecSpecific:
        CUVIDH264PICPARAMS  h264

    ctypedef struct CUVIDPICPARAMS:
        int PicWidthInMbs                           #IN: Coded frame size in macroblocks
        int FrameHeightInMbs                        #IN: Coded frame height in macroblocks
        int CurrPicIdx                              #IN: Output index of the current picture
        int field_pic_flag                          #IN: 0=frame picture, 1=field picture
        int bottom_field_flag                       #IN: 0=top field, 1=bottom field (ignored if field_pic_flag=0)
        int second_field                            #IN: Second field of a complementary field pair
        #Bitstream data
        unsigned int nBitstreamDataLen              #IN: Number of bytes in bitstream data buffer
        const unsigned char *pBitstreamData         #IN: Ptr to bitstream data for this picture (slice-layer)
        unsigned int nNumSlices                     #IN: Number of slices in this picture
        const unsigned int *pSliceDataOffsets       #IN: nNumSlices entries, contains offset of each slice within
                                                    #the bitstream data buffer
        int ref_pic_flag                            #IN: This picture is a reference picture
        int intra_pic_flag                          #IN: This picture is entirely intra coded
        unsigned int Reserved[30]                   #Reserved for future use
        #IN: Codec-specific data in union
        CUVIDCodecSpecific CodecSpecific

    ctypedef struct CUVIDPROCPARAMS:
        int progressive_frame                       #IN: Input is progressive (deinterlace_mode will be ignored)
        int second_field                            #IN: Output the second field (ignored if deinterlace mode is Weave)
        int top_field_first                         #IN: Input frame is top field first (1st field is top, 2nd field is bottom)
        int unpaired_field                          #IN: Input only contains one field (2nd field is invalid)
        #The fields below are used for raw YUV input
        unsigned int reserved_flags                 #Reserved for future use (set to zero)
        unsigned int reserved_zero                  #Reserved (set to zero)
        unsigned long long raw_input_dptr           #IN: Input CUdeviceptr for raw YUV extensions
        unsigned int raw_input_pitch                #IN: pitch in bytes of raw YUV input (should be aligned appropriately)
        unsigned int raw_input_format               #IN: Input YUV format (cudaVideoCodec_enum)
        unsigned long long raw_output_dptr          #IN: Output CUdeviceptr for raw YUV extensions
        unsigned int raw_output_pitch               #IN: pitch in bytes of raw YUV output (should be aligned appropriately)
        unsigned int Reserved1                      #Reserved for future use (set to zero)
        CUstream output_stream                      #IN: stream object used by cuvidMapVideoFrame
        unsigned int Reserved[46]                   #Reserved for future use (set to zero)
        unsigned long long *histogram_dptr          #OUT: Output CUdeviceptr for histogram extensions
        void *Reserved2[1]                          #Reserved for future use (set to zero)

    ctypedef struct CUVIDGETDECODESTATUS:
        cuvidDecodeStatus decodeStatus
        unsigned int reserved[31]
        void *pReserved[8]

    CUresult cuvidGetDecoderCaps(CUVIDDECODECAPS *pdc)
    CUresult cuvidCreateDecoder(CUvideodecoder *phDecoder, CUVIDDECODECREATEINFO *pdci)
    CUresult cuvidDestroyDecoder(CUvideodecoder hDecoder)
    CUresult cuvidDecodePicture(CUvideodecoder hDecoder, CUVIDPICPARAMS *pPicParams) nogil
    CUresult cuvidGetDecodeStatus(CUvideodecoder hDecoder, int nPicIdx, CUVIDGETDECODESTATUS* pDecodeStatus)

    #CUresult cuvidMapVideoFrame(CUvideodecoder hDecoder, int nPicIdx,
    #                            unsigned int *pDevPtr, unsigned int *pPitch,
    #                            CUVIDPROCPARAMS *pVPP)
    #CUresult CUDAAPI cuvidUnmapVideoFrame(CUvideodecoder hDecoder, unsigned int DevPtr)
    CUresult cuvidMapVideoFrame64(CUvideodecoder hDecoder, int nPicIdx, unsigned long long *pDevPtr,
                                             unsigned int *pPitch, CUVIDPROCPARAMS *pVPP)
    CUresult cuvidUnmapVideoFrame64(CUvideodecoder hDecoder, unsigned long long DevPtr)

    #we don't use threads, so no need for this:
    CUresult cuvidCtxLockCreate(CUvideoctxlock *pLock, CUcontext ctx)
    CUresult cuvidCtxLockDestroy(CUvideoctxlock lck)
    CUresult cuvidCtxLock(CUvideoctxlock lck, unsigned int reserved_flags)
    CUresult cuvidCtxUnlock(CUvideoctxlock lck, unsigned int reserved_flags)


DECODE_STATUS_STR: Dict[int, str] = {
    cuvidDecodeStatus_Invalid       : "invalid",
    cuvidDecodeStatus_InProgress    : "in-progress",
    cuvidDecodeStatus_Success       : "success",
    cuvidDecodeStatus_Error         : "error",
    cuvidDecodeStatus_Error_Concealed : "error-concealed",
}

CODEC_NAMES: Dict[int, str] = {
#    cudaVideoCodec_MPEG1    : "mpeg1",
#    cudaVideoCodec_MPEG2    : "mpeg2",
    cudaVideoCodec_MPEG4    : "mpeg4",
    cudaVideoCodec_VC1      : "vc1",
    cudaVideoCodec_H264     : "h264",
    cudaVideoCodec_JPEG     : "jpeg",
    cudaVideoCodec_H264_SVC : "h264-svc",
    cudaVideoCodec_H264_MVC : "h264-mvc",
    cudaVideoCodec_HEVC     : "hevc",
    cudaVideoCodec_VP8      : "vp8",
    cudaVideoCodec_VP9      : "vp9",
    cudaVideoCodec_AV1      : "av1",
}

CHROMA_NAMES: Dict[int, str] = {
    cudaVideoChromaFormat_Monochrome    : "monochrome",
    cudaVideoChromaFormat_420           : "420",
    cudaVideoChromaFormat_422           : "422",
    cudaVideoChromaFormat_444           : "444",
}

SURFACE_NAMES: Dict[int, str] = {
    cudaVideoSurfaceFormat_NV12     : "NV12",
    cudaVideoSurfaceFormat_P016     : "P016",
    cudaVideoSurfaceFormat_YUV444   : "YUV444P",
    cudaVideoSurfaceFormat_YUV444_16Bit : "YUV444P16",
}

CODEC_MAP = dict((v,k) for k,v in CODEC_NAMES.items())
CS_CHROMA: Dict[str, int] = {
    "YUV420P" : cudaVideoChromaFormat_420,
    "YUV422P" : cudaVideoChromaFormat_422,
    "YUV444P" : cudaVideoChromaFormat_444,
}


def init_module(options: dict) -> None:
    log("nvdec.init_module(%s)", options)
    from xpra.codecs.nvidia.util import has_nvidia_hardware
    if has_nvidia_hardware() is False:
        raise ImportError("no nvidia GPU device found")


def cleanup_module() -> None:
    log("nvdec.cleanup_module()")


def get_version() -> Tuple[int]:
    return (0, )


def get_type() -> str:
    return "nvdec"


def get_info() -> Dict[str, Any]:
    return {
        "version"   : get_version(),
    }


MIN_SIZES: Dict[str, Tuple[int, int]] = {}


def get_min_size(encoding: str) -> Tuple[int, int]:
    return MIN_SIZES.get(encoding, (48, 16))


#CODECS = ("jpeg", "h264", "vp8", "vp9")
CODECS = ("jpeg", )


def get_encodings() -> Sequence[str]:
    return CODECS


def get_specs() -> Sequence[VideoSpec]:
    specs: Sequence[VideoSpec] = []
    for encoding in CODECS:
        # has_lossless_mode = encoding == "vp9" and colorspace=="YUV444P"
        specs.append(VideoSpec(
                encoding=encoding, input_colorspace="YUV420P", output_colorspaces=("NV12", ),
                has_lossless_mode = False,
                codec_class=Decoder, codec_type=get_type(),
                quality=50, speed=100,
                size_efficiency=80,
                setup_cost=50,
                cpu_cost=0,
                gpu_cost=100,
                max_w=8192,
                max_h=4096,
            )
        )
    return specs


cdef int seq_cb(void *user_data, CUVIDEOFORMAT *vf) except 0:
    cdef Decoder decoder = <Decoder> decoders.get(int(<uintptr_t> user_data))
    return decoder.sequence_callback(vf)


cdef int decode_cb(void *user_data, CUVIDPICPARAMS *pp) except 0:
    cdef Decoder decoder = <Decoder> decoders.get(int(<uintptr_t> user_data))
    return decoder.decode_callback(pp)


cdef int display_cb(void *user_data, CUVIDPARSERDISPINFO *pdi) except 0:
    cdef Decoder decoder = <Decoder> decoders.get(int(<uintptr_t> user_data))
    return decoder.display_callback(pdi.picture_index, pdi.timestamp)


cdef int getop_cb(void *user_data, CUVIDOPERATINGPOINTINFO *op) except 0:
    #av1 specific, we don't care
    return 1


cdef int getseimsg_cb(void *user_data, CUVIDSEIMESSAGEINFO *seimsg) except 0:
    log(f"getseimsg_cb {seimsg.sei_message_count} sei messages")
    #void *pSEIData                  #OUT: SEI Message Data
    #CUSEIMESSAGE *pSEIMessage       #OUT: SEI Message Info
    #unsigned int sei_message_count  #OUT: SEI Message Count
    #unsigned int picIdx             #OUT: SEI Message Pic Index
    return 1


sequence = AtomicInteger()
decoders = WeakValueDictionary()


#note: the C api is event based but we want synchronous behaviour
#so we use an event to ensure that the data we feed produces an image as output
#this will need to be changed to support B-frames
cdef class Decoder:
    cdef unsigned int width
    cdef unsigned int height
    cdef unsigned long frames
    cdef unsigned long sequence
    cdef object colorspace
    cdef object encoding
    cdef CUvideodecoder context
    cdef CUvideoparser parser
    cdef CUstream stream
    cdef object event
    cdef object buffer
    cdef object image

    cdef object __weakref__

    def init_context(self, encoding: str, width: int, height: int, colorspace: str, options: typedict) -> None:
        log("nvdec.Decoder.init_context%s", (encoding, width, height, colorspace, options))
        if encoding not in CODEC_MAP:
            raise ValueError(f"invalid encoding {encoding} for nvdec")
        if colorspace not in CS_CHROMA:
            raise ValueError(f"invalid colorspace {colorspace} for nvdec")
        self.sequence = sequence.increase()
        self.encoding = encoding
        self.colorspace = colorspace
        self.width = width
        self.height = height
        self.event = Event()        #set each time the data has been parsed / processed
        self.stream = <CUstream> 0
        self.buffer = None
        self.image = None
        if self.encoding=="jpeg":
            self.init_decoder()
        else:
            self.init_parser()
        decoders[self.sequence] = self

    def init_parser(self) -> None:
        log("init_parser()")
        cdef CUVIDPARSERPARAMS pp
        memset(&pp, 0, sizeof(CUVIDPARSERPARAMS))
        pp.CodecType = CODEC_MAP[self.encoding]
        pp.ulMaxNumDecodeSurfaces = 0
        #pp.ulClockRate                #IN: Timestamp units in Hz (0=default=10000000Hz)
        pp.ulErrorThreshold = 100
        pp.ulMaxDisplayDelay = 0        #0=no delay
        pp.pUserData = <void *> (<uintptr_t> self.sequence)
        pp.pfnSequenceCallback = seq_cb
        pp.pfnDecodePicture = decode_cb
        pp.pfnDisplayPicture = display_cb
        pp.pfnGetOperatingPoint = getop_cb
        pp.pfnGetSEIMsg = getseimsg_cb
        cdef CUresult r
        r = cuvidCreateVideoParser(&self.parser, &pp)
        log(f"cuvidCreateVideoParser()={r}")
        cudacheck(r, "creating parser returned error")

    cdef sequence_callback(self, CUVIDEOFORMAT *vf):
        encoding = CODEC_NAMES.get(vf.codec, vf.codec)
        log("sequence_callback codec=%s", encoding)
        #log(" frame_rate=%s", vf.frame_rate)
        log(" progressive_sequence=%s", vf.progressive_sequence)
        log(" bit_depth_luma_minus8=%s, bit_depth_chroma_minus8=%s",
            vf.bit_depth_luma_minus8, vf.bit_depth_chroma_minus8)
        log(" min_num_decode_surfaces=%s, bitrate=%s", vf.min_num_decode_surfaces, vf.bitrate)
        log(" coded size: %sx%s", vf.coded_width, vf.coded_height)
        #print(" display_area=%s", vf.display_area)
        log(" chroma_format=%s", CHROMA_NAMES.get(vf.chroma_format, vf.chroma_format))
        if encoding!=self.encoding:
            log.error(f"Error: expected {self.encoding!r} but parser found {encoding!r}")
            return 0
        self.init_decoder(vf.min_num_decode_surfaces)
        self.event.set()
        return vf.min_num_decode_surfaces

    cdef int decode_callback(self, CUVIDPICPARAMS *pp):
        log("decode_callback(..)")
        self.decode_data(pp)
        return 1

    cdef int display_callback(self, int picture_index, CUvideotimestamp timestamp):
        log("display_callback(%s, %s)", picture_index, timestamp)
        self.image = self.get_output_image(picture_index)
        return int(self.image is not None)

    def init_decoder(self, int decode_surfaces=2) -> None:
        log(f"init_decoder({decode_surfaces})")
        cdef CUVIDDECODECREATEINFO pdci
        memset(&pdci, 0, sizeof(CUVIDDECODECREATEINFO))
        pdci.CodecType = CODEC_MAP[self.encoding]
        pdci.ChromaFormat = CS_CHROMA[self.colorspace]
        pdci.OutputFormat = cudaVideoSurfaceFormat_NV12 #cudaVideoSurfaceFormat_YUV444_16Bit
        pdci.bitDepthMinus8 = 0
        pdci.DeinterlaceMode = cudaVideoDeinterlaceMode_Weave
        #cudaVideoCreate_Default
        pdci.ulCreationFlags = cudaVideoCreate_PreferCUVID  #Use dedicated video engines directly
        pdci.ulIntraDecodeOnly = 0
        pdci.ulNumDecodeSurfaces = decode_surfaces
        pdci.ulNumOutputSurfaces = 1
        #geometry:
        pdci.ulWidth = self.width
        pdci.ulHeight = self.height
        pdci.ulTargetWidth = self.width
        pdci.ulTargetHeight = self.height
        pdci.ulMaxWidth = self.width
        pdci.ulMaxHeight = self.height
        #pdci.display_area.left = 0
        #pdci.display_area.top = 0
        #pdci.display_area.right = self.width
        #pdci.display_area.bottom = self.height
        #pdci.target_rect.left = 0
        #pdci.target_rect.top = 0
        #pdci.target_rect.right = self.width
        #pdci.target_rect.bottom = self.height
        pdci.vidLock = NULL
        pdci.enableHistogram = 0
        cdef CUresult r = cuvidCreateDecoder(&self.context, &pdci)
        log(f"cuvidCreateDecoder(..)={r}")
        cudacheck(r, "creating nvdec decoder returned error")

    def __repr__(self):
        return f"nvdec({self.encoding})"

    def get_info(self) -> Dict[str, Any]:
        return {
            "type"      : self.get_type(),
            "width"     : self.width,
            "height"    : self.height,
            "encoding"  : self.encoding,
            "frames"    : int(self.frames),
            "colorspace": self.colorspace,
        }

    def get_colorspace(self) -> str:
        return self.colorspace

    def get_width(self) -> int:
        return self.width

    def get_height(self) -> int:
        return self.height

    def is_closed(self) -> bool:
        return self.context==NULL

    def get_encoding(self) -> str:
        return self.encoding

    def get_type(self) -> str:
        return "nvdec"

    def __dealloc__(self):
        self.clean()

    def clean(self) -> None:
        cdef CUresult r = 0
        e = self.event
        if e:
            e.set()
            self.event = None
        if self.parser!=NULL:
            self.flush()
            r = cuvidDestroyVideoParser(self.parser)
            if r:
                log.error(f"Error destroying parser: {get_error_name(r)}")
            self.parser = NULL
        if self.context!=NULL:
            r = cuvidDestroyDecoder(self.context)
            if r:
                log.error(f"Error destroying decoder: {get_error_name(r)}")
            self.context = NULL
        self.width = 0
        self.height = 0
        self.colorspace = ""
        self.encoding = ""

    def decompress_image(self, data, options: typedict) -> ImageWrapper:
        log(f"nvdec.decompress_image({len(data)} bytes, {options})")
        cdef CUresult r
        cdef CUVIDSOURCEDATAPACKET packet
        cdef CUVIDPICPARAMS pic
        try:
            self.image = None
            stream = options.get("stream", None)
            if stream:
                self.stream = <CUstream> (<uintptr_t> stream.handle)
            with buffer_context(data) as bc:
                self.buffer = bc
                if self.parser:
                    self.event.clear()
                    memset(&packet, 0, sizeof(CUVIDSOURCEDATAPACKET))
                    packet.flags = CUVID_PKT_ENDOFPICTURE   #| CUVID_PKT_TIMESTAMP
                    packet.timestamp = 0
                    packet.payload_size = len(bc)
                    packet.payload = <const unsigned char*> (<uintptr_t> int(bc))
                    r = cuvidParseVideoData(self.parser, &packet)
                    log(f"cuvidParseVideoData(..)={r}")
                    cudacheck(r, "parsing error")
                    if not self.event.wait(1):
                        raise RuntimeError("parsing timed out")
                    return self.image
                else:
                    #no need: just use a blank pic params:
                    memset(&pic, 0, sizeof(CUVIDPICPARAMS))
                    self.decode_data(&pic)
                    return self.get_output_image()
        finally:
            self.buffer = None
            self.stream = <CUstream> 0

    cdef decode_data(self, CUVIDPICPARAMS *pic):
        log(f"decode_data({len(self.buffer)} bytes)")
        self.frames += 1
        cdef CUresult r = 0
        pic.nBitstreamDataLen = len(self.buffer)
        pic.pBitstreamData = <const unsigned char*> (<uintptr_t> int(self.buffer))
        with nogil:
            r = cuvidDecodePicture(self.context, pic)
        if r==CUDA_ERROR_INVALID_IMAGE:
            raise ValueError("invalid image data")
        log(f"cuvidDecodePicture()={r}")
        cudacheck(r, "GPU picture decoding returned error")
        cdef CUVIDGETDECODESTATUS status
        memset(&status, 0, sizeof(CUVIDGETDECODESTATUS))
        cdef int pic_idx = pic.CurrPicIdx
        #start = monotonic()
        r = cuvidGetDecodeStatus(self.context, pic_idx, &status)
        #while status.decodeStatus==cuvidDecodeStatus_InProgress:
        #    sleep(0.001)
        #    r = cuvidGetDecodeStatus(self.context, pic_idx, &status)
        #    log("waiting for decode")
        #end = monotonic()
        sinfo = DECODE_STATUS_STR.get(status.decodeStatus, status.decodeStatus)
        log(f"decompress_image: status={sinfo}")
        #if status.decodeStatus not in (
        #    cuvidDecodeStatus_InProgress,
        #    cuvidDecodeStatus_Success,
        #    cuvidDecodeStatus_Error_Concealed,
        #    ):
        #    if r in DECODE_STATUS_STR:
        #        raise RuntimeError(f"GPU decoding status returned error {sinfo!r}")
        #    cudacheck(r, "GPU picture decoding status error")

    def get_output_image(self, int pic_idx=0) -> ImageWrapper:
        #map it as a CUDA buffer:
        cdef CUVIDPROCPARAMS map_params
        map_params.progressive_frame = 1
        memset(&map_params, 0, sizeof(CUVIDPROCPARAMS))
        map_params.output_stream = self.stream
        cdef unsigned long long dev_ptr
        cdef unsigned int pitch
        r = cuvidMapVideoFrame64(self.context, pic_idx, &dev_ptr, &pitch, &map_params)
        cudacheck(r, "GPU mapping of picture buffer error")
        log(f"mapped picture {pic_idx} at {dev_ptr:x}, pitch={pitch}, stream=%s", <uintptr_t> self.stream)
        try:
            yuv_buf, yuv_pitch = mem_alloc_pitch(self.width, roundup(self.height, 2)*3//2, 4)
            copy = Memcpy2D()
            copy.set_src_device(dev_ptr)
            copy.src_x_in_bytes = 0
            copy.src_y = 0
            copy.src_pitch = pitch
            copy.width_in_bytes = self.width
            copy.height = roundup(self.height, 2)*3//2
            copy.set_dst_device(yuv_buf)
            copy.dst_x_in_bytes = 0
            copy.dst_y = 0
            copy.dst_pitch = yuv_pitch
            copy(aligned=False)
            rowstrides = [yuv_pitch, yuv_pitch]
            image = ImageWrapper(0, 0, self.width, self.height,
                                 yuv_buf, "NV12", 24, rowstrides, 3, ImageWrapper.PLANAR_2)
            self.frames += 1
            self.event.set()
            return image
        finally:
            r = cuvidUnmapVideoFrame64(self.context, dev_ptr)
            cudacheck(r, "error unmapping video frame")

    cdef flush(self):
        cdef CUVIDSOURCEDATAPACKET packet
        cdef CUresult r
        if self.parser:
            memset(&packet, 0, sizeof(CUVIDSOURCEDATAPACKET))
            r = cuvidParseVideoData(self.parser, &packet)
            packet.payload_size = 0
            packet.payload = NULL
            packet.flags = CUVID_PKT_ENDOFSTREAM
            r = cuvidParseVideoData(self.parser, &packet)
            log(f"cuvidParseVideoData(..)={r}")


def download_from_gpu(buf, size_t size) -> MemBuf:
    log("nvdec download_from_gpu%s", (buf, size))
    start = monotonic()
    cdef MemBuf pixels = getbuf(size, False)
    memcpy_dtoh(pixels, buf)
    end = monotonic()
    log("nvdec downloaded %i bytes in %ims", size, 1000*(end-start))
    return pixels


def decompress(encoding: str, img_data, width: int, height: int, options=None) -> ImageWrapper:
    #decompress using the default device,
    #and download the pixel data from the GPU:
    dev = get_default_device_context()
    if not dev:
        raise RuntimeError("no cuda device found")
    with dev as cuda_context:
        log("cuda_context=%s for device=%s", cuda_context, dev.get_info())
        options = typedict(options or {})
        stream = options.get("stream")
        if not stream:
            stream = Stream()
            options["stream"] = stream
        return decompress_and_download(encoding, img_data, width, height, options)


def decompress_and_download(encoding: str, img_data, width: int, height: int, options=None) -> ImageWrapper:
    img = decompress_with_device(encoding, img_data, width, height, options)
    cuda_buffer = img.get_pixels()
    rowstrides = img.get_rowstride()
    assert len(rowstrides)==2 and rowstrides[0]==rowstrides[1]
    stride = rowstrides[0]
    assert img.get_pixel_format()=="NV12"
    height = img.get_height()
    size = roundup(height, 2)*stride*3//2
    y_size = stride*height
    uv_start = roundup(height, 2)*stride
    uv_size = stride*(height//2)
    nv12_buf = download_from_gpu(cuda_buffer, size)
    pixels = memoryview(nv12_buf)
    planes = pixels[:y_size], pixels[uv_start:uv_start+uv_size]
    cuda_buffer.free()
    img.set_pixels(planes)
    return img


def decompress_with_device(encoding: str, img_data, width: int, height: int, options=None) -> ImageWrapper:
    cdef Decoder decoder = Decoder()
    try:
        tdoptions = typedict(options or {})
        decoder.init_context(encoding, width, height, "YUV420P", tdoptions)
        return decoder.decompress_image(img_data, tdoptions)
    finally:
        decoder.clean()


def selftest(full=False) -> None:
    from xpra.codecs.nvidia.util import has_nvidia_hardware, get_nvidia_module_version
    if not has_nvidia_hardware():
        raise ImportError("no nvidia GPU device found")
    get_nvidia_module_version(True)

    dev = get_default_device_context()
    if not dev:
        raise RuntimeError("no device found")

    cdef CUVIDDECODECAPS caps
    cdef CUresult r
    cdef Decoder decoder

    codec_ok = {}
    codec_failed = []
    with dev as cuda_context:
        log("cuda_context=%s for device=%s", cuda_context, dev.get_info())

        for codec_i, codec_name in CODEC_NAMES.items():
            min_w = min_h = 0
            chroma_ok= []
            chroma_failed = []
            for chroma_i, chroma_name in CHROMA_NAMES.items():
                memset(&caps, 0, sizeof(CUVIDDECODECAPS))
                caps.eCodecType = codec_i
                caps.eChromaFormat = chroma_i
                caps.nBitDepthMinus8 = 0
                r = cuvidGetDecoderCaps(&caps)
                if r:
                    chroma_failed.append(chroma_name)
                    log(f"decoder caps for {codec_name} + {chroma_name} returned error %s", get_error_name(r))
                    continue
                if not caps.bIsSupported:
                    chroma_failed.append(chroma_name)
                    #log(f"{codec_name} + {chroma_name} is not supported on this GPU")
                    continue
                if caps.nMaxWidth<4096 or caps.nMaxHeight<4096:
                    chroma_failed.append(chroma_name)
                    log(f"{codec_name} maximum dimension is only {caps.nMaxWidth}x{caps.nMaxHeight}")
                    continue
                oformats = tuple(name for sfi, name in SURFACE_NAMES.items() if caps.nOutputFormatMask & (1<<sfi))
                log(f"output formats for {codec_name} + {chroma_name}: %s (mask={caps.nOutputFormatMask:x})", csv(oformats))
                if "NV12" in oformats:
                    if min_w==0 or caps.nMinWidth>min_w:
                        min_w = caps.nMinWidth
                    if min_h==0 or caps.nMinHeight>min_h:
                        min_h = caps.nMinHeight
                    chroma_ok.append(chroma_name)
                else:
                    log(f"{codec_name} does not support NV12 surface")
            if chroma_ok:
                codec_ok[codec_name] = chroma_ok
                MIN_SIZES[codec_name] = (min_w, min_h)
            else:
                codec_failed.append(codec_name)
        log(f"codecs failed: {codec_failed}")
        log(f"codecs supported: {codec_ok}")
        log(f"minimum sizes: {MIN_SIZES}")
        from xpra.codecs.checks import testdecoder
        from xpra.codecs.nvidia import nvdec
        nvdec.decoder.CODECS = testdecoder(nvdec.decoder, full)
        log(f"validated encodings: {nvdec.decoder.CODECS}")
        if not nvdec.decoder.CODECS:
            raise RuntimeError("no encodings supported")



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2016 Antoine Martin <antoine@xpra.org>

import gi
gi.require_version('Gdk', '3.0')  # @UndefinedVariable
gi.require_version('Gtk', '3.0')  # @UndefinedVariable
